@@ -1,4 +1,4 @@
-# Copyright (C) 2022-2024 Hibiki AI Limited <info@hibiki-ai.com>
+# Copyright (C) 2022-2025 Hibiki AI Limited <info@hibiki-ai.com>
 #
 # This file is part of mirai.
 #
@@ -20,7 +20,7 @@
 #'
 #' Set \sQuote{daemons} or persistent background processes to receive
 #' \code{\link{mirai}} requests. Specify \sQuote{n} to create daemons on the
-#' local machine. Specify \sQuote{url} for receiving connections from remote
+#' local machine. Specify \sQuote{url} to receive connections from remote
 #' daemons (for distributed computing across the network). Specify
 #' \sQuote{remote} to optionally launch remote daemons via a remote
 #' configuration. By default, dispatcher ensures optimal scheduling.
@@ -43,36 +43,32 @@
 #' a task, it will exit as soon as the task is complete.
 #'
 #' To reset persistent daemons started with \code{autoexit = FALSE}, use
-#' \code{daemons(NULL)} instead, which also sends exit instructions to all
-#' connected daemons prior to resetting.
+#' \code{daemons(NULL)} instead, which also sends exit signals to all connected
+#' daemons prior to resetting.
 #'
 #' For historical reasons, \code{daemons()} with no arguments returns the value
 #' of \code{\link{status}}.
 #'
 #' @inheritParams mirai
 #' @inheritParams dispatcher
-#' @param n integer number of daemons to set.
-#' @param url [default NULL] if specified, the character URL or vector of URLs
-#'   on the host for remote daemons to dial into, including a port accepting
-#'   incoming connections (and optionally for websockets, a path), e.g.
-#'   'tcp://hostname:5555' or 'ws://10.75.32.70:5555/path'. Specify a URL
-#'   starting 'tls+tcp://' or 'wss://' to use secure TLS connections. Auxiliary
-#'   function \code{\link{host_url}} may be used to construct a valid host URL.
+#' @param n integer number of daemons to launch.
+#' @param url [default NULL] if specified, a character string comprising a URL
+#'   at which to listen for remote daemons, including a port accepting incoming
+#'   connections, e.g. 'tcp://hostname:5555' or 'tcp://10.75.32.70:5555'.
+#'   Specify a URL with scheme 'tls+tcp://' to use secure TLS connections (for
+#'   details see Distributed Computing section below). Auxiliary function
+#'   \code{\link{host_url}} may be used to construct a valid host URL.
 #' @param remote [default NULL] required only for launching remote daemons, a
 #'   configuration generated by \code{\link{remote_config}} or
 #'   \code{\link{ssh_config}}.
-#' @param dispatcher [default 'process'] character value, one of
-#'   \sQuote{process}, \sQuote{thread} or \sQuote{none}. Whether to deploy
-#'   dispatcher in another process, on a thread or not at all. Dispatcher is an
-#'   extension that ensures optimal scheduling, although this is not always
-#'   required (for details see Dispatcher section below). Note that the option
-#'   \sQuote{thread} is new and currently considered experimental.
+#' @param dispatcher [default TRUE] logical value, whether to use dispatcher.
+#'   Dispatcher runs in a separate process to ensure optimal scheduling,
+#'   although this may not always be required (for details see Dispatcher
+#'   section below).
 #' @param ... (optional) additional arguments passed through to
-#'   \code{\link{dispatcher}} if using dispatcher and/or \code{\link{daemon}} if
-#'   launching daemons. These include \sQuote{retry} and \sQuote{token} at
-#'   dispatcher and \sQuote{asyncdial}, \sQuote{autoexit}, \sQuote{cleanup},
-#'   \sQuote{output}, \sQuote{maxtasks}, \sQuote{idletime}, \sQuote{walltime}
-#'   and \sQuote{timerstart} at daemon.
+#'   \code{\link{daemon}} if launching daemons. These include \sQuote{asyncdial},
+#'   \sQuote{autoexit}, \sQuote{cleanup}, \sQuote{output}, \sQuote{maxtasks},
+#'   \sQuote{idletime} and \sQuote{walltime}.
 #' @param force [default TRUE] logical value whether to always reset daemons and
 #'   apply new settings for a compute profile, even if already set. If FALSE,
 #'   applying new settings requires daemons to be explicitly reset first using
@@ -84,6 +80,10 @@
 #'   of task completions would otherwise influence the tasks sent to each
 #'   daemon. Even for \code{dispatcher = 'none'}, reproducibility is not
 #'   guaranteed if the order in which tasks are sent is not deterministic.
+#' @param serial [default NULL] (optional, requires dispatcher) a configuration
+#'   created by \code{\link{serial_config}} to register serialization and
+#'   unserialization functions for normally non-exportable reference objects,
+#'   such as Arrow Tables or torch tensors.
 #' @param tls [default NULL] (optional for secure TLS connections) if not
 #'   supplied, zero-configuration single-use keys and certificates are
 #'   automatically generated. If supplied, \strong{either} the character path to
@@ -93,9 +93,8 @@
 #'   comprising [i] the TLS certificate (optionally certificate chain) and [ii]
 #'   the associated private key.
 #'
-#' @return If using dispatcher, the integer number of daemons set, or else the
-#'   integer number of daemons launched locally (zero if using a remote
-#'   launcher).
+#' @return The integer number of daemons launched locally (zero if specifying
+#'   \sQuote{url} / using a remote launcher).
 #'
 #' @section Local Daemons:
 #'
@@ -109,15 +108,11 @@
 #'
 #' @section Dispatcher:
 #'
-#' By default \code{dispatcher = "process"} launches a background process
-#' running \code{\link{dispatcher}}. Dispatcher connects to daemons on behalf of
+#' By default \code{dispatcher = TRUE} launches a background process running
+#' \code{\link{dispatcher}}. Dispatcher connects to daemons on behalf of
 #' the host and ensures optimal FIFO scheduling of tasks.
 #'
-#' Specifying \code{dispatcher = "thread"} runs dispatcher logic on a new
-#' thread, a faster and more efficient alternative to using a separate process.
-#' This is a new feature and should be considered experimental.
-#'
-#' Specifying \code{dispatcher = "none"}, uses the default behaviour without
+#' Specifying \code{dispatcher = FALSE}, uses the default behaviour without
 #' additional dispatcher logic. In this case daemons connect directly to the
 #' host and tasks are distributed in a round-robin fashion. Optimal scheduling
 #' is not guaranteed as the duration of tasks cannot be known \emph{a priori},
@@ -128,17 +123,20 @@
 #'
 #' @section Distributed Computing:
 #'
-#' Specifying \sQuote{url} allows tasks to be distributed across the network.
-#' This should be a character string such as \sQuote{tcp://10.75.32.70:5555} at
-#' which daemon processes should connect to. Switching the URL scheme to
-#' \sQuote{tls+tcp://} or \sQuote{wss://} automatically upgrades the connection
-#' to use TLS. The auxiliary function \code{\link{host_url}} may be used to
-#' automatically construct a valid host URL based on the computer's hostname.
+#' Specifying \sQuote{url} as a character string allows tasks to be distributed
+#' across the network. \sQuote{n} is not required in this case, and disregarded
+#' if supplied.
 #'
-#' Specify \sQuote{remote} with a call to \code{\link{remote_config}} or
-#' \code{\link{ssh_config}} to launch daemons on remote machines. Otherwise,
-#' \code{\link{launch_remote}} may be used to generate the shell commands to
-#' deploy daemons manually on remote resources.
+#' Supply a URL with a \sQuote{tcp://} scheme, such as
+#' \sQuote{tcp://10.75.32.70:5555}. The host / dispatcher listens at this
+#' address, utilising a single port. Individual daemons (started with
+#' \code{\link{daemon}}) may then dial in to this URL. Host / dispatcher
+#' automatically adjusts to the number of daemons actually connected, allowing
+#' dynamic upscaling or downscaling as required.
+#'
+#' Switching the URL scheme to \sQuote{tls+tcp://} automatically upgrades the
+#' connection to use TLS. The auxiliary function \code{\link{host_url}} may be
+#' used to construct a valid host URL based on the computer's hostname.
 #'
 #' IPv6 addresses are also supported and must be enclosed in square brackets [ ]
 #' to avoid confusion with the final colon separating the port. For example,
@@ -146,51 +144,13 @@
 #' \sQuote{tcp://[::1]:5555}.
 #'
 #' Specifying the wildcard value zero for the port number e.g.
-#' \sQuote{tcp://[::1]:0} or \sQuote{ws://[::1]:0} will automatically assign a
-#' free ephemeral port. Use \code{\link{status}} to inspect the actual assigned
-#' port at any time.
+#' \sQuote{tcp://[::1]:0} will automatically assign a free ephemeral port. Use
+#' \code{\link{status}} to inspect the actual assigned port at any time.
 #'
-#' \strong{With Dispatcher}
-#'
-#' When using dispatcher, it is recommended to use a websocket URL rather than
-#' TCP, as this requires only one port to connect to all daemons: a websocket
-#' URL supports a path after the port number, which can be made unique for each
-#' daemon.
-#'
-#' Specifying a single host URL such as \sQuote{ws://10.75.32.70:5555} with
-#' \code{n = 6} will automatically append a sequence to the path, listening to
-#' the URLs \sQuote{ws://10.75.32.70:5555/1} through
-#' \sQuote{ws://110.75.32.70:5555/6}.
-#'
-#' Alternatively, specify a vector of URLs to listen to arbitrary port numbers /
-#' paths. In this case it is optional to supply \sQuote{n} as this can be
-#' inferred by the length of vector supplied.
-#'
-#' Individual daemons then dial in to each of these host URLs. At most one
-#' daemon can be dialled into each URL at any given time.
-#'
-#' Dispatcher automatically adjusts to the number of daemons actually connected.
-#' Hence it is possible to dynamically scale up or down the number of daemons as
-#' required, subject to the maximum number initially specified.
-#'
-#' Alternatively, supplying a single TCP URL will listen at a block of URLs with
-#' ports starting from the supplied port number and incrementing by one for
-#' \sQuote{n} specified e.g. the host URL \sQuote{tcp://10.75.32.70:5555} with
-#' \code{n = 6} listens to the contiguous block of ports 5555 through 5560.
-#'
-#' \strong{Without Dispatcher}
-#'
-#' A TCP URL may be used in this case as the host listens at only one address,
-#' utilising a single port.
-#'
-#' The network topology is such that daemons (started with \code{\link{daemon}})
-#' or indeed dispatchers (started with \code{\link{dispatcher}}) dial into the
-#' same host URL.
-#'
-#' \sQuote{n} is not required in this case, and disregarded if supplied, as
-#' network resources may be added or removed at any time. The host automatically
-#' distributes tasks to all connected daemons and dispatchers in a round-robin
-#' fashion.
+#' Specify \sQuote{remote} with a call to \code{\link{remote_config}} or
+#' \code{\link{ssh_config}} to launch daemons on remote machines. Otherwise,
+#' \code{\link{launch_remote}} may be used to generate the shell commands to
+#' deploy daemons manually on remote resources.
 #'
 #' @section Compute Profiles:
 #'
@@ -228,19 +188,19 @@
 #' daemons(0)
 #'
 #' # Create 2 local daemons (not using dispatcher)
-#' daemons(2, dispatcher = "none")
+#' daemons(2, dispatcher = FALSE)
 #' status()
 #' # Reset to zero
 #' daemons(0)
 #'
-#' # 2 remote daemons via dispatcher using WebSockets
-#' daemons(2, url = host_url(ws = TRUE))
+#' # Set up dispatcher accepting TLS over TCP connections
+#' daemons(url = host_url(tls = TRUE))
 #' status()
 #' # Reset to zero
 #' daemons(0)
 #'
 #' # Set host URL for remote daemons to dial into
-#' daemons(url = host_url(), dispatcher = "none")
+#' daemons(url = host_url(), dispatcher = FALSE)
 #' status()
 #' # Reset to zero
 #' daemons(0)
@@ -251,33 +211,35 @@
 #'   {
 #'     m1 <- mirai(Sys.getpid())
 #'     m2 <- mirai(Sys.getpid())
-#'     cat(call_mirai(m1)$data, call_mirai(m2)$data, "\n")
+#'     cat(m1[], m2[], "\n")
 #'   }
 #' )
 #'
 #' }
 #'
 #' \dontrun{
-#' # Launch 2 daemons on remotes 'nodeone' and 'nodetwo' using SSH
+#' # Launch daemons on remotes 'nodeone' and 'nodetwo' using SSH
 #' # connecting back directly to the host URL over a TLS connection:
 #'
-#' daemons(url = host_url(tls = TRUE),
+#' daemons(n = 1L,
+#'         url = host_url(tls = TRUE),
 #'         remote = ssh_config(c('ssh://nodeone', 'ssh://nodetwo')),
-#'         dispatcher = "none")
+#'         dispatcher = FALSE)
 #'
 #' # Launch 4 daemons on the remote machine 10.75.32.90 using SSH tunnelling
-#' # over port 5555 ('url' hostname must be 'localhost' or '127.0.0.1'):
+#' # over port 5555 ('url' hostname must be '127.0.0.1'):
 #'
-#' daemons(n = 4,
-#'         url = 'ws://localhost:5555',
-#'         remote = ssh_config('ssh://10.75.32.90', tunnel = TRUE))
+#' daemons(n = 4L,
+#'         url = 'tcp://127.0.0.1:5555',
+#'         remote = ssh_config('ssh://10.75.32.90', tunnel = TRUE, port = 5555))
 #'
 #' }
 #'
 #' @export
 #'
-daemons <- function(n, url = NULL, remote = NULL, dispatcher = c("process", "thread", "none"),
-                    ..., force = TRUE, seed = NULL, tls = NULL, pass = NULL, .compute = "default") {
+daemons <- function(n, url = NULL, remote = NULL, dispatcher = TRUE, ...,
+                    force = TRUE, seed = NULL, serial = NULL,
+                    tls = NULL, pass = NULL, .compute = "default") {
 
   missing(n) && missing(url) && return(status(.compute))
 
@@ -286,44 +248,49 @@ daemons <- function(n, url = NULL, remote = NULL, dispatcher = c("process", "thr
   if (is.character(url)) {
 
     if (is.null(envir)) {
-      envir <- init_envir_stream(seed = seed)
-      tls <- check_create_tls(url = url, tls = tls, envir = envir)
+      envir <- init_envir_stream(seed)
+      launches <- 0L
+      dots <- parse_dots(...)
+      output <- attr(dots, "output")
       switch(
-        parse_dispatcher(dispatcher[1L]),
+        parse_dispatcher(dispatcher),
+        {
+          tls <- configure_tls(url, tls, pass, envir)
+          sock <- req_socket(url, tls = tls)
+          check_store_url(sock, envir)
+        },
+        {
+          tls <- configure_tls(url, tls, pass, envir, returnconfig = FALSE)
+          cv <- cv()
+          urld <- local_url()
+          sock <- req_socket(urld)
+          res <- launch_sync_dispatcher(sock, sock, wa52(urld, dots, url), output, tls, pass, serial)
+          is.object(res) && stop(._[["sync_dispatcher"]])
+          store_dispatcher(sock, res, cv, envir)
+          `[[<-`(envir, "msgid", 0L)
+        },
         {
           n <- if (missing(n)) length(url) else if (is.numeric(n) && n >= 1L) as.integer(n) else stop(._[["n_one"]])
-          if (length(tls)) tls_config(server = tls, pass = pass)
+          tls <- configure_tls(url, tls, pass, envir, returnconfig = FALSE)
           cv <- cv()
-          dots <- parse_dots(...)
-          output <- attr(dots, "output")
           urld <- local_url()
           urlc <- sprintf("%s%s", urld, "c")
           sock <- req_socket(urld)
           sockc <- req_socket(urlc)
           res <- launch_sync_dispatcher(sock, sockc, wa5(urld, dots, n, urlc, url), output, tls, pass)
           is.object(res) && stop(._[["sync_dispatcher"]])
-          store_dispatcher(sockc = sockc, res = res, cv = cv, envir = envir)
-        },
-        {
-          n <- if (missing(n)) length(url) else if (is.numeric(n) && n >= 1L) as.integer(n) else stop(._[["n_one"]])
-          urls <- resolve_dispatcher_urls(n = n, url = url)
-          sock <- .dispatcher(host = inproc_url(), url = urls, tls = if (length(tls)) tls_config(server = tls, pass = pass))
-          `[[<-`(`[[<-`(envir, "urls", urls), "dispatcher", TRUE)
-        },
-        {
-          sock <- req_socket(url, tls = if (length(tls)) tls_config(server = tls, pass = pass))
-          check_store_url(sock = sock, envir = envir)
-          n <- 0L
+          store_dispatcher(sockc, res, cv, envir)
+          launches <- n
         },
         stop(._[["dispatcher_args"]])
       )
-      `[[<-`(.., .compute, `[[<-`(`[[<-`(envir, "sock", sock), "n", n))
+      `[[<-`(.., .compute, `[[<-`(`[[<-`(`[[<-`(envir, "sock", sock), "n", launches), "dots", dots))
       if (length(remote))
-        launch_remote(url = envir[["urls"]], remote = remote, tls = envir[["tls"]], ..., .compute = .compute)
-    } else {
-      daemons(n = 0L, .compute = .compute)
+        launch_remote(n = n, remote = remote, tls = envir[["tls"]], ..., .compute = .compute)
+    } else if (force) {
+      daemons(0L, .compute = .compute)
       return(daemons(n = n, url = url, remote = remote, dispatcher = dispatcher, ...,
-                     seed = seed, tls = tls, pass = pass, .compute = .compute))
+                     seed = seed, serial = serial, tls = tls, pass = pass, .compute = .compute))
     }
 
   } else {
@@ -336,7 +303,7 @@ daemons <- function(n, url = NULL, remote = NULL, dispatcher = c("process", "thr
     if (n == 0L) {
       is.null(envir) && return(0L)
 
-      if (signal) send_signal(envir = envir)
+      if (signal) send_signal(envir)
       reap(envir[["sock"]])
       is.null(envir[["sockc"]]) || reap(envir[["sockc"]])
       ..[[.compute]] <- NULL -> envir
@@ -344,12 +311,26 @@ daemons <- function(n, url = NULL, remote = NULL, dispatcher = c("process", "thr
     } else if (is.null(envir)) {
 
       n > 0L || stop(._[["n_zero"]])
-      envir <- init_envir_stream(seed = seed)
+      envir <- init_envir_stream(seed)
       urld <- local_url()
       dots <- parse_dots(...)
       output <- attr(dots, "output")
       switch(
-        parse_dispatcher(dispatcher[1L]),
+        parse_dispatcher(dispatcher),
+        {
+          sock <- req_socket(urld)
+          launch_sync_daemons(seq_len(n), sock, urld, dots, envir, output) || stop(._[["sync_daemons"]])
+          `[[<-`(envir, "urls", urld)
+        },
+        {
+          cv <- cv()
+          sock <- req_socket(urld)
+          res <- launch_sync_dispatcher(sock, sock, wa42(urld, dots, envir[["stream"]], n), output, serial = serial)
+          is.object(res) && stop(._[["sync_dispatcher"]])
+          store_dispatcher(sock, res, cv, envir)
+          for (i in seq_len(n)) next_stream(envir)
+          `[[<-`(envir, "msgid", 0L)
+        },
         {
           cv <- cv()
           sock <- req_socket(urld)
@@ -357,28 +338,16 @@ daemons <- function(n, url = NULL, remote = NULL, dispatcher = c("process", "thr
           sockc <- req_socket(urlc)
           res <- launch_sync_dispatcher(sock, sockc, wa4(urld, dots, envir[["stream"]], n, urlc), output)
           is.object(res) && stop(._[["sync_dispatcher"]])
-          store_dispatcher(sockc = sockc, res = res, cv = cv, envir = envir)
+          store_dispatcher(sockc, res, cv, envir)
           for (i in seq_len(n)) next_stream(envir)
         },
-        {
-          urls <- auto_dispatcher_urls(n = n, url = urld)
-          sock <- .dispatcher(host = inproc_url(), url = urls)
-          for (i in seq_len(n))
-            launch_daemon(wa3(urls[i], dots, next_stream(envir)), output)
-          `[[<-`(`[[<-`(envir, "urls", urls), "dispatcher", TRUE)
-        },
-        {
-          sock <- req_socket(urld)
-          launch_sync_daemons(seq_len(n), sock, urld, dots, envir, output) || stop(._[["sync_daemons"]])
-          `[[<-`(envir, "urls", urld)
-        },
         stop(._[["dispatcher_args"]])
       )
-      `[[<-`(.., .compute, `[[<-`(`[[<-`(envir, "sock", sock), "n", n))
+      `[[<-`(.., .compute, `[[<-`(`[[<-`(`[[<-`(envir, "sock", sock), "n", n), "dots", dots))
     } else if (force) {
-      daemons(n = 0L, .compute = .compute)
+      daemons(0L, .compute = .compute)
       return(daemons(n = n, url = url, remote = remote, dispatcher = dispatcher, ...,
-                     seed = seed, tls = tls, pass = pass, .compute = .compute))
+                     seed = seed, serial = serial, tls = tls, pass = pass, .compute = .compute))
     }
 
   }
@@ -395,7 +364,9 @@ print.miraiDaemons <- function(x, ...) print(unclass(x))
 #' With Mirai Daemons
 #'
 #' Evaluate an expression with daemons that last for the duration of the
-#' expression.
+#' expression. Ensure each mirai within the statement is explicitly called (or
+#' their values collected) so that daemons are not reset before they have all
+#' completed.
 #'
 #' This function is an S3 method for the generic \code{with} for class
 #' 'miraiDaemons'.
@@ -411,11 +382,11 @@ print.miraiDaemons <- function(x, ...) print(unclass(x))
 #' # Only run examples in interactive R sessions
 #'
 #' with(
-#'   daemons(2),
+#'   daemons(2, dispatcher = FALSE),
 #'   {
 #'     m1 <- mirai(Sys.getpid())
 #'     m2 <- mirai(Sys.getpid())
-#'     cat(call_mirai(m1)$data, call_mirai(m2)$data, "\n")
+#'     cat(m1[], m2[], "\n")
 #'   }
 #' )
 #'
@@ -444,45 +415,29 @@ with.miraiDaemons <- function(data, expr, ...) {
 #'
 #' @return A named list comprising:
 #'   \itemize{
-#'     \item \strong{connections} - integer number of active connections.
-#'     \cr Using dispatcher: Always 1L as there is a single connection to
-#'     dispatcher, which connects to the daemons in turn.
-#'     \item \strong{daemons} - of variable type.
-#'     \cr Using dispatcher: a status matrix (see Status Matrix section below),
-#'     or else an integer \sQuote{errorValue} if communication with dispatcher
-#'     failed.
-#'     \cr Not using dispatcher: the character host URL.
-#'     \cr Not set: 0L.
+#'     \item \strong{connections} - integer number of active daemon connections.
+#'     \item \strong{daemons} - character URL at which host / dispatcher is
+#'     listening, or else \code{0L} if daemons have not yet been set.
+#'     \item \strong{mirai} (present only if using dispatcher) - a named integer
+#'     vector comprising: \strong{awaiting} - number of tasks queued for
+#'     execution at dispatcher, \strong{executing} - number of tasks sent to a
+#'     daemon for execution, and \strong{completed} - number of tasks for which
+#'     the result has been received (either completed or cancelled).
 #'   }
 #'
-#' @section Status Matrix:
+#' @section Events:
 #'
-#' When using dispatcher, \code{$daemons} comprises an integer matrix with the
-#' following columns:
-#'
-#' \itemize{
-#'   \item \strong{i} - integer index number.
-#'   \item \strong{online} - shows as 1 when there is an active connection, or
-#'   else 0 if a daemon has yet to connect or has disconnected.
-#'   \item \strong{instance} - increments by 1 every time there is a new
-#'   connection at a URL. This counter is designed to track new daemon instances
-#'   connecting after previous ones have ended (due to time-outs etc.). The
-#'   count becomes negative immediately after a URL is regenerated by
-#'   \code{\link{saisei}}, but increments again once a new daemon connects.
-#'   \item \strong{assigned} - shows the cumulative number of tasks assigned to
-#'   the daemon.
-#'   \item \strong{complete} - shows the cumulative number of tasks completed by
-#'   the daemon.
-#' }
-#'
-#' The dispatcher URLs are stored as row names to the matrix.
+#'   If dispatcher is used combined with daemon IDs, an additional element
+#'   \strong{events} will report the positive integer ID when the daemon
+#'   connects and the negative value when it disconnects. Only the events since
+#'   the previous status query are returned.
 #'
 #' @examples
 #' if (interactive()) {
 #' # Only run examples in interactive R sessions
 #'
 #' status()
-#' daemons(n = 2L, url = "wss://[::1]:0")
+#' daemons(url = "tcp://[::1]:0")
 #' status()
 #' daemons(0)
 #'
@@ -494,10 +449,8 @@ status <- function(.compute = "default") {
 
   is.list(.compute) && return(status(attr(.compute, "id")))
   envir <- ..[[.compute]]
-  is.null(envir) &&
-    return(list(connections = 0L, daemons = 0L))
-  is.null(envir[["dispatcher"]]) ||
-    return(dispatcher_status(envir))
+  is.null(envir) && return(list(connections = 0L, daemons = 0L))
+  length(envir[["msgid"]]) && return(dispatcher_status(envir))
   list(connections = as.integer(stat(envir[["sock"]], "pipes")),
        daemons = if (is.null(envir[["sockc"]])) envir[["urls"]] else query_status(envir))
 
@@ -527,7 +480,7 @@ status <- function(.compute = "default") {
 #'   \code{torch::torch_load}.
 #'
 #' @return A list comprising the configuration. This should be passed to the
-#'   \sQuote{.serial} argument of \code{\link{everywhere}}.
+#'   \sQuote{serial} argument of \code{\link{daemons}}.
 #'
 #' @examples
 #' cfg <- serial_config("test_cls", function(x) serialize(x, NULL), unserialize)
@@ -539,9 +492,7 @@ serial_config <- serial_config
 
 # internals --------------------------------------------------------------------
 
-inproc_url <- function() sprintf("inproc://%s", random(10L))
-
-check_create_tls <- function(url, tls, envir) {
+configure_tls <- function(url, tls, pass, envir, returnconfig = TRUE) {
   purl <- parse_url(url)
   sch <- purl[["scheme"]]
   if ((startsWith(sch, "wss") || startsWith(sch, "tls")) && is.null(tls)) {
@@ -549,7 +500,9 @@ check_create_tls <- function(url, tls, envir) {
     `[[<-`(envir, "tls", cert[["client"]])
     tls <- cert[["server"]]
   }
-  tls
+  cfg <- if (length(tls)) tls_config(server = tls, pass = pass)
+  returnconfig || return(tls)
+  cfg
 }
 
 init_envir_stream <- function(seed) {
@@ -562,53 +515,22 @@ init_envir_stream <- function(seed) {
   envir
 }
 
-# "%s-%d" format required by IPC under MacOS
-auto_dispatcher_urls <- function(n, url)
-  as.character(
-    lapply(
-      seq_len(n),
-      function(x) sprintf(if (startsWith(url, "ipc")) "%s-%d" else "%s/%d", url, x)
-    )
-  )
-
-resolve_url_port <- function(url) {
-  parse_url(url)[["port"]] == "0" || return(url)
-  sock <- socket(listen = url)
-  port <- opt(attr(sock, "listener")[[1L]], "tcp-bound-port")
-  reap(sock)
-  sub_real_port(port = port, url = url)
-}
-
-resolve_dispatcher_urls <- function(n, url) {
-  for (i in seq_along(url))
-    url[[i]] <- resolve_url_port(url[[i]])
-  n == length(url) && return(url)
-  if (length(url) == 1L && n > 1L && startsWith(url, "t")) {
-    port <- parse_url(url)[["port"]]
-    ports <- seq.int(from = port, length.out = n)
-    as.character(lapply(ports, sub, pattern = port, x = url, fixed = TRUE))
-  } else {
-    auto_dispatcher_urls(n = n, url = url)
-  }
-}
-
 tokenized_url <- function(url) sprintf("%s/%s", url, random(12L))
 
 req_socket <- function(url, tls = NULL, resend = 0L)
-  `opt<-`(socket(protocol = "req", listen = url, tls = tls), "req:resend-time", resend)
+  `opt<-`(socket("req", listen = url, tls = tls), "req:resend-time", resend)
 
 parse_dispatcher <- function(x)
-  if (x == "process") 1L else if (x == "thread") 2L else if (x == "none") 3L else
-    if (is.logical(x)) 1L + (!x) * 2L else 4L
+  if (is.logical(x)) 1L + (!is.na(x) && x) else if (x == "process" || x == "thread") 3L else if (x == "none") 1L else 4L
 
 parse_dots <- function(...) {
   ...length() || return("")
   dots <- list(...)
   dots <- dots[as.logical(lapply(dots, function(x) is.logical(x) || is.numeric(x)))]
+  length(dots) || return("")
   dnames <- names(dots)
   out <- sprintf(",%s", paste(dnames, dots, sep = "=", collapse = ","))
-  pos <- dnames == "output"
-  any(pos) && as.logical(dots[pos])[1L] && return(`attr<-`(out, "output", ""))
+  is.logical(dots[["output"]]) && dots[["output"]] && return(`attr<-`(out, "output", ""))
   out
 }
 
@@ -617,27 +539,36 @@ parse_tls <- function(tls)
 
 libp <- function(lp = .libPaths()) lp[file.exists(file.path(lp, "mirai"))][1L]
 
-wa3 <- function(url, dots, rs, tls = NULL)
+wa31 <- function(url, dots, rs, tls = NULL)
   shQuote(sprintf("mirai::daemon(\"%s\"%s%s,rs=c(%s))", url, dots, parse_tls(tls), paste0(rs, collapse = ",")))
 
+wa3 <- function(url, dots, rs, tls = NULL)
+  shQuote(sprintf("mirai::daemon(\"%s\"%s%s,rs=c(%s),dispatcher=FALSE)", url, dots, parse_tls(tls), paste0(rs, collapse = ",")))
+
+wa32 <- function(url, dots, rs, tls = NULL)
+  shQuote(sprintf("mirai::daemon(\"%s\"%s%s,rs=c(%s),dispatcher=TRUE)", url, dots, parse_tls(tls), paste0(rs, collapse = ",")))
+
 wa4 <- function(urld, dots, rs, n, urlc)
   shQuote(sprintf(".libPaths(c(\"%s\",.libPaths()));mirai::dispatcher(\"%s\",n=%d,rs=c(%s),monitor=\"%s\"%s)", libp(), urld, n, paste0(rs, collapse= ","), urlc, dots))
 
+wa42 <- function(urld, dots, rs, n)
+  shQuote(sprintf(".libPaths(c(\"%s\",.libPaths()));mirai::dispatcher(\"%s\",n=%d,rs=c(%s)%s)", libp(), urld, n, paste0(rs, collapse= ","), dots))
+
 wa5 <- function(urld, dots, n, urlc, url)
   shQuote(sprintf(".libPaths(c(\"%s\",.libPaths()));mirai::dispatcher(\"%s\",c(\"%s\"),n=%d,monitor=\"%s\"%s)", libp(), urld, paste0(url, collapse = "\",\""), n, urlc, dots))
 
-launch_daemon <- function(args, output)
-  system2(command = .command, args = c("-e", args), stdout = output, stderr = output, wait = FALSE)
+wa52 <- function(urld, dots, url)
+  shQuote(sprintf(".libPaths(c(\"%s\",.libPaths()));mirai::dispatcher(\"%s\",url=\"%s\"%s)", libp(), urld, url, dots))
 
-launch_sync_dispatcher <- function(sock, sockc, args, output, tls = NULL, pass = NULL) {
+launch_daemon <- function(args, output)
+  system2(.command, args = c("-e", args), stdout = output, stderr = output, wait = FALSE)
+
+launch_sync_dispatcher <- function(sock, sockc, args, output, tls = NULL, pass = NULL, serial = NULL) {
   pkgs <- Sys.getenv("R_DEFAULT_PACKAGES")
-  system2(command = .command, args = c("--default-packages=NULL", "--vanilla", "-e", args), stdout = output, stderr = output, wait = FALSE)
-  vec <- c("p", pkgs,
-           "o", if (is.character(tls)) tls[1L] else "",
-           "s", if (is.character(tls) && length(tls) > 1L) tls[2L] else "",
-           "i", if (is.character(pass)) pass[1L] else "",
-           "t")
-  query_dispatcher(sockc, command = vec, mode = 2L, block = .limit_long)
+  system2(.command, args = c("--default-packages=NULL", "--vanilla", "-e", args), stdout = output, stderr = output, wait = FALSE)
+  if (is.list(serial))
+    `opt<-`(sock, "serial", serial)
+  query_dispatcher(sockc, list(pkgs, tls, pass, serial), send_mode = 1L, recv_mode = 2L, block = .limit_long)
 }
 
 launch_sync_daemons <- function(seq, sock, urld, dots, envir, output) {
@@ -657,21 +588,20 @@ check_store_url <- function(sock, envir) {
   listener <- attr(sock, "listener")[[1L]]
   url <- opt(listener, "url")
   if (parse_url(url)[["port"]] == "0")
-    url <- sub_real_port(port = opt(listener, "tcp-bound-port"), url = url)
+    url <- sub_real_port(opt(listener, "tcp-bound-port"), url)
   `[[<-`(envir, "urls", url)
 }
 
 send_signal <- function(envir) {
   signals <- max(length(envir[["urls"]]), stat(envir[["sock"]], "pipes"))
   for (i in seq_len(signals)) {
-    send(envir[["sock"]], data = ._scm_., mode = 2L)
+    send(envir[["sock"]], ._scm_., mode = 2L)
     msleep(10L)
   }
 }
 
 query_status <- function(envir) {
-  res <- query_dispatcher(sock = envir[["sockc"]], command = 0L, mode = 5L)
-  is.object(res) && return(res)
+  res <- query_dispatcher(envir[["sockc"]], 0L)
   `attributes<-`(
     res,
     list(
@@ -682,11 +612,15 @@ query_status <- function(envir) {
 }
 
 dispatcher_status <- function(envir) {
-  online <- .online(envir[["sock"]])
-  list(
-    connections = sum(online),
-    daemons = `dimnames<-`(`dim<-`(online, c(envir[["n"]], 1L)), list(envir[["urls"]], "online"))
-  )
+  status <- query_dispatcher(envir[["sock"]], c(0L, 0L))
+  out <- list(connections = status[1L],
+              daemons = envir[["urls"]],
+              mirai = c(awaiting = status[2L],
+                        executing = status[3L],
+                        completed = envir[["msgid"]] - status[2L] - status[3L]))
+  if (length(status) > 3L)
+    out <- c(out, list(events = status[4:length(status)]))
+  out
 }
 
 ._scm_. <- as.raw(c(0x42, 0x0a, 0x03, 0x00, 0x00, 0x00, 0x02, 0x03, 0x04, 0x00, 0x00, 0x05, 0x03, 0x00, 0x05, 0x00, 0x00, 0x00, 0x55, 0x54, 0x46, 0x2d, 0x38, 0xfc, 0x00, 0x00, 0x00))
