@@ -1,4 +1,4 @@
-# Copyright (C) 2022-2024 Hibiki AI Limited <info@hibiki-ai.com>
+# Copyright (C) 2022-2025 Hibiki AI Limited <info@hibiki-ai.com>
 #
 # This file is part of mirai.
 #
@@ -45,11 +45,11 @@
 #' @author Charlie Gao \email{charlie.gao@@shikokuchuo.net}
 #'   (\href{https://orcid.org/0000-0002-0750-061X}{ORCID})
 #'
-#' @importFrom nanonext .advance collect_aio collect_aio_ call_aio call_aio_
-#'   .context cv cv_value dial .dispatcher is_error_value .keep listen lock
-#'   .mark mclock msleep nng_error .online opt opt<- parse_url pipe_notify
-#'   random reap recv recv_aio request send serial_config socket stat stop_aio
-#'   tls_config unresolved .unresolved until wait write_cert
+#' @importFrom nanonext .advance call_aio call_aio_ collect_aio collect_aio_
+#'   .context cv cv_signal cv_value dial .interrupt is_error_value .keep listen
+#'   lock .mark mclock monitor msleep nng_error opt opt<- parse_url pipe_notify
+#'   random read_monitor reap recv recv_aio request send serial_config socket
+#'   stat stop_aio tls_config unresolved .unresolved until wait write_cert
 #'
 "_PACKAGE"
 
@@ -85,15 +85,14 @@
 
 ._ <- list2env(
   list(
-    arglen = "'args' and/or 'url' must be of length 1 or the same length",
+    arglen = "'n' must equal the length of 'args', or either must be 1",
     cluster_inactive = "cluster is no longer active",
-    correct_context = "'host' must be specified if not using directly in a function argument",
     daemons_unset = "daemons must be set to use launchers",
-    dispatcher_args = "'dispatcher' must be one of 'process', 'thread' or 'none'",
+    dispatcher_args = "'dispatcher' must be either TRUE or FALSE",
     dot_required = "'.' must be an element of the character vector(s) supplied to 'args'",
     function_required = "'.f' must be of type function, not %s",
     missing_expression = "missing expression, perhaps wrap in {}?",
-    missing_url = "at least one URL must be supplied for 'url' or 'n' must be at least 1",
+    missing_url = "'n' must be 1 or greater, or else 'url' must be supplied",
     named_args = "all items in '.args' must be named, unless supplying an environment",
     named_dots = "all '...' arguments must be named, unless supplying an environment",
     n_one = "'n' must be 1 or greater",
@@ -102,11 +101,8 @@
     numeric_n = "'n' must be numeric, did you mean to provide 'url'?",
     register_cluster = "this function requires a more recent version of R",
     requires_daemons = "mirai is launching one local daemon for a map operation as none previously set",
-    requires_local = "SSH tunnelling requires 'url' / 'host' to be '127.0.0.1' or 'localhost'",
-    single_url = "only one 'url' should be specified",
     sync_daemons = "initial sync with daemon(s) timed out after 10s",
-    sync_dispatcher = "initial sync with dispatcher timed out after 10s",
-    url_spec = "numeric value for 'url' is out of bounds"
+    sync_dispatcher = "initial sync with dispatcher timed out after 10s"
   ),
   hash = TRUE
 )
