@@ -24,26 +24,23 @@ test_error(mirai(a, .args = list(1)), "all items in '.args' must be named")
 test_error(mirai_map(1:2, "a function"), "must be of type function, not character")
 test_error(daemons(url = "URL"), "Invalid argument")
 test_error(daemons(-1), "zero or greater")
-test_error(daemons(n = 0, url = "ws://localhost:0"), "1 or greater")
 test_error(daemons(raw(0L)), "must be numeric")
-test_error(daemons(1, dispatcher = "p"), "must be one of")
-test_error(daemons(url = local_url(), dispatcher = "t"), "must be one of")
-test_error(dispatcher(client = "URL"), "at least one")
+test_error(daemons(1, dispatcher = "p"), "must be either")
+test_error(daemons(url = local_url(), dispatcher = "t"), "must be either")
+test_error(dispatcher(client = "URL"), "must be 1 or greater")
 test_error(daemon("URL"), "Invalid argument")
 test_error(launch_local(1L), "daemons must be set")
 test_type("character", host_url())
-test_true(startsWith(host_url(ws = TRUE, tls = TRUE), "wss"))
 test_true(startsWith(host_url(tls = TRUE), "tls"))
 test_true(grepl("5555", host_url(port = 5555), fixed = TRUE))
 test_type("list", ssh_config("ssh://remotehost"))
-test_type("list", ssh_config("ssh://remotehost", tunnel = TRUE, host = "tls+tcp://127.0.0.1:5555"))
-test_error(ssh_config("ssh://remotehost", tunnel = TRUE), "'host' must be specified")
+test_type("list", ssh_config("ssh://remotehost", tunnel = TRUE, port = "5555"))
 test_true(is_mirai_interrupt(r <- mirai:::mk_interrupt_error()))
 test_print(r)
 test_true(is_mirai_error(r <- `class<-`("Error in: testing\n", c("miraiError", "errorValue", "try-error"))))
 test_print(r)
 test_null(r$stack.trace)
-test_equal(mirai:::.DollarNames.miraiError(NULL, "s"), "stack.trace")
+test_equal(mirai:::.DollarNames.miraiError(r, "c"), "class")
 test_true(mirai:::is.promising.mirai())
 test_error(everywhere({}), "not found")
 test_null(nextstream())
@@ -64,7 +61,7 @@ connection && {
   args <- c(m = 2L, n = 4L)
   m <- mirai(.expr = `lang obj`, .args = args, .timeout = 2000L)
   if (!is_error_value(call_mirai_(m)$data)) test_equal(m$data, 8L)
-  test_null(stop_mirai(m))
+  test_true(!stop_mirai(m))
   Sys.sleep(1L)
   test_equal(1L, d <- daemons(1L, dispatcher = FALSE, asyncdial = FALSE, seed = 1546L))
   test_print(d)
@@ -84,18 +81,18 @@ connection && {
   test_print(dm)
   test_type("integer", status()[["connections"]])
   test_type("character", status()[["daemons"]])
-  test_type("character", mlc <- launch_remote("ws://[::1]:5555"))
+  test_type("character", mlc <- launch_remote())
   test_class("miraiLaunchCmd", mlc)
   test_print(mlc)
-  test_error(launch_remote("ws://[::1]:5555", remote = remote_config(command = "echo", args = "invalid")), "must be an element")
-  test_error(launch_remote(c("tcp://localhost:5555", "tcp://localhost:6666", "tcp://localhost:7777"), remote = remote_config(command = "echo", args = list(c("test", "."), c("test", ".")))), "must be of length 1 or the same length")
+  test_error(launch_remote(1L, remote = remote_config(command = "echo", args = "invalid")), "must be an element")
+  test_error(launch_remote(3L, remote = remote_config(command = "echo", args = list(c("test", "."), c("test", ".")))), "must equal the length")
   test_zero(daemons(0L))
   Sys.sleep(1L)
-  test_equal(1L, daemons(1L, dispatcher = FALSE, idletime = 500L, timerstart = 1L, cleanup = FALSE, output = TRUE, .compute = "new"))
+  test_equal(1L, daemons(1L, dispatcher = FALSE, maxtasks = 10L, walltime = 10000L, idletime = 20000L, cleanup = FALSE, output = TRUE, .compute = "new"))
   test_type("character", nextget("urls", .compute = "new"))
   test_type("integer", nextstream(.compute = "new"))
   Sys.sleep(1.5)
-  test_null(everywhere({}, as.environment(df), .compute = "new"))
+  test_type("list", everywhere({}, as.environment(df), .compute = "new"))
   mn <- mirai("test1", .compute = "new")
   mp <- mirai(b + 1, .compute = "new")
   Sys.sleep(1L)
@@ -108,16 +105,19 @@ connection && {
 # additional daemons tests
 connection && .Platform[["OS.type"]] != "windows" && {
   Sys.sleep(1L)
-  test_zero(daemons(url = value <- local_url(), dispatcher = "none"))
+  test_zero(daemons(url = value <- local_url(), dispatcher = FALSE))
   test_identical(status()$daemons, value)
   test_identical(nextget("urls"), value)
-  test_type("character", launch_remote("ws://[::1]:5555", remote = remote_config(command = "echo", args = list(c("Test out:", ".", ">/dev/null")), rscript = "/usr/lib/R/bin/Rscript")))
-  test_type("character", launch_remote("tcp://localhost:5555", remote = ssh_config(remotes = c("ssh://remotehost", "ssh://remotenode"), tunnel = TRUE, command = "echo")))
+  test_type("character", launch_remote(remote = remote_config(command = "echo", args = list(c("Test out:", ".", ">/dev/null")), rscript = "/usr/lib/R/bin/Rscript")))
+  test_type("character", launch_remote(remote = ssh_config(remotes = c("ssh://remotehost", "ssh://remotenode"), tunnel = TRUE, port = "5555", command = "echo")))
   test_zero(daemons(0L))
   Sys.sleep(1L)
   test_zero(daemons(n = 2L, url = value <- "ws://:0", dispatcher = FALSE, remote = remote_config(quote = TRUE)))
   test_true(status()$daemons != value)
   test_zero(daemons(0L))
+}
+# mirai_map tests
+connection && .Platform[["OS.type"]] != "windows" && {
   Sys.sleep(1L)
   m <- with(daemons(1, dispatcher = "none", .compute = "ml"), {
     if (is.null(tryCatch(mirai_map(list(1, "a", 2), sum, .compute = "ml")[.stop], error = function(e) NULL)))
@@ -133,6 +133,7 @@ connection && .Platform[["OS.type"]] != "windows" && {
   test_identical(call_mirai(mp)[["x"]][["data"]], "ab")
   test_true(all(mirai_map(data.frame(1:3, 3:1), sum, .args = list(3L))[.flat] == 7L))
   test_true(all(mirai_map(list(c(a = 1, b = 1, c = 1), 3), sum)[.flat] == 3))
+  test_zero(daemons(0L))
 }
 # parallel cluster tests
 library(parallel)
@@ -204,113 +205,156 @@ connection && {
 # advanced daemons and dispatcher tests
 connection && .Platform[["OS.type"]] != "windows" && Sys.getenv("NOT_CRAN") == "true" && {
   Sys.sleep(1L)
-  test_equal(1L, daemons(url = local_url(), dispatcher = "process", notused = "wrongtype"))
-  test_true(grepl("://", launch_remote(1L), fixed = TRUE))
-  test_null(launch_local(nextget("urls")))
-  Sys.sleep(1L)
-  requireNamespace("promises", quietly = TRUE) && {
-    test_true(promises::is.promise(p1 <- promises::as.promise(mirai("completed"))))
-    test_true(promises::is.promise(p2 <- promises::`%...>%`(mirai("completed"), identity())))
-    test_true(promises::is.promise(p3 <- promises::as.promise(call_mirai(mirai("completed")))))
-    test_zero(mirai_map(0:1, function(x) x, .promise = identity)[][[1L]])
-    test_true(is_mirai_map(mp <- mirai_map(matrix(1:4, nrow = 2L), function(x, y) x + y, .promise = list(identity))))
-    test_true(all(mp[.flat, .stop] == c(4L, 6L)))
-    test_null(names(mp[]))
-    test_class("errorValue", mirai_map(1, function(x) stop(x), .promise = list(identity, identity))[][[1L]])
-  }
-  Sys.sleep(1L)
-  test_zero(daemons(NULL))
-  test_equal(1L, daemons(url = "ws://:0", correctype = 0L, token = TRUE))
+  test_zero(daemons(url = "ws://:0", correctype = 0L, token = TRUE))
   test_zero(daemons(0L))
   test_zero(with(daemons(url = "tcp://:0", correcttype = c(1, 0), token = TRUE), {8L - 9L + 1L}))
-  test_equal(daemons(n = 2, "ws://:0"), 2L)
+  test_zero(daemons(n = 2, "ws://:0"))
   test_type("integer", nextget("pid"))
-  test_equal(length(nextget("urls")), 2L)
+  test_equal(length(nextget("urls")), 1L)
   Sys.sleep(1L)
-  status <- status()[["daemons"]]
-  test_class("matrix", status)
-  test_type("character", dn1 <- dimnames(status)[[1L]])
-  test_type("character", parse1 <- nanonext::parse_url(dn1[1L]))
-  test_type("character", parse2 <- nanonext::parse_url(dn1[2L]))
-  test_true((port <- as.integer(parse1[["port"]])) > 0L)
-  test_equal(as.integer(parse2[["port"]]), port)
-  test_equal(parse1[["path"]], "/1")
-  test_equal(parse2[["path"]], "/2")
-  test_zero(sum(status[, "online"]))
-  test_zero(sum(status[, "instance"]))
-  test_zero(sum(status[, "assigned"]))
-  test_zero(sum(status[, "complete"]))
-  test_type("character", saisei(i = 1L))
-  test_null(saisei(i = 0L))
-  test_type("character", saisei(i = 1L, force = TRUE))
-  test_null(saisei(i = 10L))
+  status <- daemons()
+  test_type("list", status)
+  test_zero(status[["connections"]])
+  test_type("integer", status[["mirai"]])
+  test_zero(status[["mirai"]][["awaiting"]])
+  test_zero(status[["mirai"]][["executing"]])
+  test_zero(status[["mirai"]][["completed"]])
   test_zero(daemons(0))
-  test_equal(daemons(url = c("tcp://127.0.0.1:45555", "tcp://127.0.0.1:0"), correcttype = NA), 2L)
-  Sys.sleep(1L)
-  test_null(launch_local(nextget("urls", .compute = "default")[1L], maxtasks = 1L))
-  test_null(launch_local(2, maxtasks = 1L))
-  Sys.sleep(2L)
-  tstatus <- status()[["daemons"]]
-  test_class("matrix", tstatus)
-  test_type("character", tdn1 <- dimnames(tstatus)[[1L]])
-  test_type("character", tparse1 <- nanonext::parse_url(tdn1[1L]))
-  test_type("character", tparse2 <- nanonext::parse_url(tdn1[2L]))
-  test_true(tparse1[["port"]] != tparse2[["port"]])
-  test_equal(sum(tstatus[, "online"]), 2L)
-  test_equal(sum(tstatus[, "instance"]), 2L)
-  test_zero(sum(tstatus[, "assigned"]))
-  test_zero(sum(tstatus[, "complete"]))
+  test_equal(daemons(2, correcttype = NA), 2L)
+  test_equal(daemons()[["connections"]], 2L)
   test_type("list", res <- mirai_map(c(1,1), rnorm)[.progress])
   test_true(res[[1L]] != res[[2L]])
-  test_zero(daemons(0))
-  test_equal(1L, daemons(url = "wss://127.0.0.1:0", token = TRUE, pass = "test"))
-  test_null(launch_local(1L))
+  test_equal(2L, daemons()[["connections"]])
+  test_zero(daemons(0L))
+  Sys.sleep(1L)
+  test_zero(daemons(url = "tls+tcp://127.0.0.1:0", dispatcher = TRUE))
+  test_equal(launch_local(1L), 1L)
   Sys.sleep(1L)
   test_true(grepl("CERTIFICATE", launch_remote(1L), fixed = TRUE))
   q <- quote(list2env(list(b = 2), envir = .GlobalEnv))
+  m <- mirai("Seattle", .timeout = 1000)
+  if (!is_error_value(m[])) test_equal(m[], "Seattle")
+  test_class("errorValue", mirai(q(), .timeout = 1000)[])
+  test_zero(daemons(0))
+}
+# TLS tests
+connection && Sys.getenv("NOT_CRAN") == "true" && {
+  Sys.sleep(1L)
   cfg <- serial_config("custom", function(x) serialize(x, NULL), unserialize)
-  test_null(everywhere(q, .serial = cfg))
+  test_zero(daemons(url = "tls+tcp://127.0.0.1:0", pass = "test", serial = cfg))
+  test_equal(launch_local(1L), 1L)
+  Sys.sleep(1L)
+  test_true(grepl("CERTIFICATE", launch_remote(1L), fixed = TRUE))
+  q <- quote(list2env(list(b = 2), envir = .GlobalEnv))
+  test_type("list", everywhere(q))
   m <- mirai(b, .timeout = 1000)
   if (!is_error_value(m[])) test_equal(m[], 2L)
   test_null(saisei(1))
-  test_error(launch_local(0:1), "out of bounds")
-  test_error(launch_remote(1:2), "out of bounds")
-  option <- 15L
-  Sys.setenv(R_DEFAULT_PACKAGES = "stats,utils")
-  test_equal(1L, daemons(1, dispatcher = TRUE, maxtasks = 10L, timerstart = 1L, walltime = 1000L, seed = 1546, token = TRUE, cleanup = option, autoexit = tools::SIGCONT))
-  Sys.unsetenv("R_DEFAULT_PACKAGES")
-  Sys.sleep(1L)
-  mq <- mirai(runif(1L), .timeout = 1000)
-  test_true(is.numeric(mq[]))
-  mq <- mirai(Sys.sleep(1.5), .timeout = 500)
-  test_class("matrix", status()[["daemons"]])
-  Sys.sleep(2L)
   test_zero(daemons(0))
-  Sys.sleep(1L)
   test_tls <- function(cert) {
     file <- tempfile()
     on.exit(unlink(file))
     cat(cert[["server"]], file = file)
-    daemons(url = "tls+tcp://127.0.0.1:0", tls = file) == 1L && daemons(0L) == 0L
+    daemons(url = "tls+tcp://127.0.0.1:0", tls = file) == 0L && daemons(0L) == 0L
   }
   test_true(test_tls(nanonext::write_cert(cn = "127.0.0.1")))
 }
-# threaded dispatcher tests
+# promises tests
+connection && requireNamespace("promises", quietly = TRUE) && Sys.getenv("NOT_CRAN") == "true" && {
+  Sys.sleep(1L)
+  test_equal(daemons(1, notused = "wrongtype"), 1L)
+  test_true(grepl("://", launch_remote(1L), fixed = TRUE))
+  test_true(promises::is.promise(p1 <- promises::as.promise(mirai("completed"))))
+  test_true(promises::is.promise(p2 <- promises::`%...>%`(mirai("completed"), identity())))
+  test_true(promises::is.promise(p3 <- promises::as.promise(call_mirai(mirai("completed")))))
+  test_zero(mirai_map(0:1, function(x) x, .promise = identity)[][[1L]])
+  test_true(is_mirai_map(mp <- mirai_map(matrix(1:4, nrow = 2L), function(x, y) x + y, .promise = list(identity))))
+  test_true(all(mp[.flat, .stop] == c(4L, 6L)))
+  test_null(names(mp[]))
+  test_class("errorValue", mirai_map(1, function(x) stop(x), .promise = list(identity, identity))[][[1L]])
+  Sys.sleep(1L)
+  getNamespace("later")[["run_now"]]()
+  test_zero(daemons(NULL))
+}
+# mirai daemon limits tests
 connection && Sys.getenv("NOT_CRAN") == "true" && {
-  test_equal(daemons(2, dispatcher = "thread"), 2L)
-  test_equal(nextget("n"), 2L)
-  test_true(startsWith(nextget("urls")[[1L]], mirai:::.urlscheme))
-  test_class("matrix", status()$daemons)
-  test_true(mirai(TRUE)[])
-  test_zero(daemons(0))
-  test_equal(daemons(2, url = host_url(), dispatcher = "thread"), 2L)
-  test_equal(length(urls <- nextget("urls")), 2L)
-  test_identical(as.integer(nanonext::parse_url(urls[[2L]])[["port"]]), as.integer(nanonext::parse_url(urls[[1L]])[["port"]]) + 1L)
-  test_zero(daemons(0))
-  test_equal(daemons(2, url = host_url(ws = TRUE), dispatcher = "thread"), 2L)
-  test_equal(length(urls <- nextget("urls")), 2L)
-  test_true(endsWith(urls[[1L]], "/1"))
-  test_true(endsWith(urls[[2L]], "/2"))
+  Sys.sleep(1L)
+  test_equal(daemons(1, cleanup = FALSE, maxtasks = 2L, id = 125L), 1L)
+  test_equal(mirai(1)[], mirai(1)[])
+  m <- mirai(0L)
+  Sys.sleep(1L)
+  res <- status()
+  test_zero(res$connections)
+  test_identical(res$events, c(125L, -125L))
+  test_equal(status()$mirai[["awaiting"]], 1L)
+  test_equal(launch_local(1, idletime = 5000L, walltime = 500L, id = 129L), 1L)
+  test_zero(m[])
+  Sys.sleep(1L)
+  res <- status()
+  test_zero(res$connections)
+  test_identical(res$events, c(129L, -129L))
   test_zero(daemons(0))
 }
+# mirai cancellation tests
+connection && Sys.getenv("NOT_CRAN") == "true" && {
+  Sys.sleep(1L)
+  test_equal(daemons(1, dispatcher = TRUE, cleanup = FALSE), 1L)
+  m1 <- mirai({ Sys.sleep(1); res <<- "m1 done" })
+  m2 <- mirai({ Sys.sleep(1); res <<- "m2 done" })
+  Sys.sleep(0.1)
+  test_true(stop_mirai(m2))
+  test_true(stop_mirai(m1))
+  test_equal(m2$data, 20L)
+  test_equal(m1$data, 20L)
+  test_class("errorValue", mirai(res)[])
+  m <- mirai_map(1:10, function(x) { Sys.sleep(2); y <<- TRUE })
+  Sys.sleep(0.1)
+  s <- stop_mirai(m)
+  test_equal(sum(unlist(m[])), 200L)
+  test_class("errorValue", mirai(y)[])
+  test_identical(s, !logical(10L))
+  test_equal(status()$connections, 1L)
+  test_equal(length(nextget("urls")), 1L)
+  test_class("miraiLaunchCmd", launch_remote(1))
+  test_zero(daemons(0))
+}
+# additional stress testing
+connection && Sys.getenv("NOT_CRAN") == "true" && {
+  Sys.sleep(1L)
+  q <- vector(mode = "list", length = 10000L)
+  Sys.setenv(R_DEFAULT_PACKAGES = "stats,utils")
+  test_equal(daemons(4), 4L)
+  Sys.unsetenv("R_DEFAULT_PACKAGES")
+  for (i in seq_len(10000L)) {q[[i]] <- mirai(1L); attr(q[[i]], "status") <- status()}
+  test_equal(sum(unlist(collect_mirai(q))), 10000L)
+  test_true(all(as.logical(lapply(lapply(q, attr, "status"), is.list))))
+  for (i in seq_len(10000L)) {q[[i]] <- mirai({Sys.sleep(0.001); rnorm(1)}); attr(q[[i]], "status") <- status()}
+  test_equal(length(unique(unlist(collect_mirai(q)))), 10000L)
+  test_true(all(as.logical(lapply(lapply(q, attr, "status"), is.list))))
+  test_equal(daemons()[["mirai"]][["completed"]], 20000L)
+}
+# legacy interface tests
+connection && .Platform[["OS.type"]] != "windows" && Sys.getenv("NOT_CRAN") == "true" && {
+  option <- 15L
+  Sys.setenv(R_DEFAULT_PACKAGES = "stats,utils")
+  test_equal(1L, daemons(1, dispatcher = "process", maxtasks = 10L, timerstart = 1L, walltime = 500L, seed = 1546, cleanup = option, autoexit = tools::SIGCONT))
+  Sys.unsetenv("R_DEFAULT_PACKAGES")
+  Sys.sleep(1L)
+  mq <- mirai(runif(1L), .timeout = 1000)
+  test_true(is.numeric(mq[]))
+  mq <- mirai(Sys.sleep(0.7), .timeout = 500)
+  test_class("matrix", status()[["daemons"]])
+  test_null(saisei(i = 1L))
+  test_type("character", saisei(i = 1L, force = TRUE))
+  Sys.sleep(0.1)
+  test_equal(daemons(url = "wss://127.0.0.1:0", dispatcher = "thread", output = TRUE, token = TRUE), 1L)
+  test_equal(nextget("n"), 1L)
+  test_equal(length(nextget("urls")), 1L)
+  test_class("matrix", status()$daemons)
+  test_null(saisei(i = 0L))
+  test_print(saisei(i = 1L))
+  Sys.sleep(0.1)
+  test_zero(daemons(0))
+}
+test_zero(daemons(0))
 Sys.sleep(1L)
